import ast
import asyncio
import json
import re
from collections.abc import Coroutine, Mapping
from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import Any, TypeVar, overload

from pydantic import TypeAdapter, ValidationError
from tqdm.autonotebook import tqdm

logger = getLogger(__name__)

_JSON_START_RE = re.compile(r"[{\[]")

T = TypeVar("T")


def extract_json_substring(text: str) -> str | None:
    decoder = json.JSONDecoder()
    for match in _JSON_START_RE.finditer(text):
        start = match.start()
        try:
            _, end = decoder.raw_decode(text, idx=start)
            return text[start:end]
        except ValueError:
            continue

    return None


def parse_json_or_py_string(
    s: str, return_none_on_failure: bool = False
) -> dict[str, Any] | list[Any] | None:
    s_fmt = re.sub(r"```[a-zA-Z0-9]*\n|```", "", s).strip()
    try:
        return ast.literal_eval(s_fmt)
    except (ValueError, SyntaxError):
        try:
            return json.loads(s_fmt)
        except json.JSONDecodeError as exc:
            if return_none_on_failure:
                return None
            raise ValueError(
                "Invalid JSON/Python string - Both ast.literal_eval and json.loads "
                f"failed to parse the following response:\n{s}"
            ) from exc


def parse_json_or_py_substring(
    json_str: str, return_none_on_failure: bool = False
) -> dict[str, Any] | list[Any] | None:
    return parse_json_or_py_string(
        extract_json_substring(json_str) or "", return_none_on_failure
    )


@overload
def validate_obj_from_json_or_py_string(
    s: str,
    adapter: TypeAdapter[T],
    from_substring: bool = False,
) -> T: ...


@overload
def validate_obj_from_json_or_py_string(
    s: str,
    adapter: Mapping[str, TypeAdapter[T]],
    from_substring: bool = False,
) -> T | str: ...


def validate_obj_from_json_or_py_string(
    s: str,
    adapter: TypeAdapter[T] | Mapping[str, TypeAdapter[T]],
    from_substring: bool = False,
) -> T | str:
    _selected_adapter: TypeAdapter[T] | None = None
    if isinstance(adapter, Mapping):
        for _marker, _adapter in adapter.items():
            if _marker in s:
                _selected_adapter = _adapter
        if _selected_adapter is None:
            return s
    else:
        _selected_adapter = adapter

    try:
        if from_substring:
            parsed = parse_json_or_py_substring(s, return_none_on_failure=True)
        else:
            parsed = parse_json_or_py_string(s, return_none_on_failure=True)
        if parsed is None:
            parsed = s
        return _selected_adapter.validate_python(parsed)
    except (json.JSONDecodeError, ValidationError) as exc:
        raise ValueError(
            f"Invalid JSON or Python string:\n{s}\n"
            f"Expected type: {_selected_adapter._type}",  # type: ignore[arg-type]
        ) from exc


def extract_xml_list(text: str) -> list[str]:
    pattern = re.compile(r"<(chunk_\d+)>(.*?)</\1>", re.DOTALL)

    chunks: list[str] = []
    for match in pattern.finditer(text):
        content = match.group(2).strip()
        chunks.append(content)
    return chunks


def read_txt(file_path: str | Path, encoding: str = "utf-8") -> str:
    return Path(file_path).read_text(encoding=encoding)


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str | bytes:
    try:
        if binary_mode:
            return Path(file_path).read_bytes()
        return Path(file_path).read_text()
    except FileNotFoundError:
        logger.error(f"File {file_path} not found.")
        return ""


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        return read_contents_from_file(prompt_path) if prompt_path is not None else None  # type: ignore[arg-type]

    return prompt_text


async def asyncio_gather_with_pbar(
    *corouts: Coroutine[Any, Any, Any],
    no_tqdm: bool = False,
    desc: str | None = None,
) -> list[Any]:
    pbar = tqdm(total=len(corouts), desc=desc, disable=no_tqdm)

    async def run_and_update(coro: Coroutine[Any, Any, Any]) -> Any:
        result = await coro
        pbar.update(1)
        return result

    wrapped_tasks = [run_and_update(c) for c in corouts]
    results = await asyncio.gather(*wrapped_tasks)
    pbar.close()

    return results


def get_timestamp() -> str:
    return datetime.now().strftime("%Y%m%d_%H%M%S")



import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Mapping, Sequence
from typing import Any, Generic, TypeVar, cast
from uuid import uuid4

from pydantic import BaseModel, TypeAdapter
from typing_extensions import TypedDict

from .memory import MessageHistory
from .typing.completion import Completion, CompletionChunk
from .typing.converters import Converters
from .typing.message import AssistantMessage, Conversation
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


class LLMSettings(TypedDict):
    pass


SettingsT = TypeVar("SettingsT", bound=LLMSettings, covariant=True)  # noqa: PLC0105
ConvertT = TypeVar("ConvertT", bound=Converters, covariant=True)  # noqa: PLC0105


class LLM(ABC, Generic[SettingsT, ConvertT]):
    @abstractmethod
    def __init__(
        self,
        converters: ConvertT,
        model_name: str | None = None,
        model_id: str | None = None,
        llm_settings: SettingsT | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: type | Mapping[str, type] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()

        self._converters = converters
        self._model_id = model_id or str(uuid4())[:8]
        self._model_name = model_name
        self._tools = {t.name: t for t in tools} if tools else None
        self._llm_settings: SettingsT = llm_settings or cast("SettingsT", {})

        self._response_format = response_format
        self._response_format_pyd: (
            TypeAdapter[Any] | Mapping[str, TypeAdapter[Any]] | None
        )
        if isinstance(response_format, type):
            self._response_format_pyd = TypeAdapter(response_format)
        elif isinstance(response_format, Mapping):
            self._response_format_pyd = {
                k: TypeAdapter(v) for k, v in response_format.items()
            }
        else:
            self._response_format_pyd = None

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def model_name(self) -> str | None:
        return self._model_name

    @property
    def llm_settings(self) -> SettingsT:
        return self._llm_settings

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, Any]] | None:
        return self._tools

    @property
    def response_format(self) -> type | Mapping[str, type] | None:
        return self._response_format

    @tools.setter
    def tools(self, tools: list[BaseTool[BaseModel, Any, Any]] | None) -> None:
        self._tools = {t.name: t for t in tools} if tools else None

    @response_format.setter
    def response_format(self, response_format: type | None) -> None:
        self._response_format = response_format

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(model_id={self.model_id}; "
            f"model_name={self._model_name})"
        )

    @abstractmethod
    async def generate_completion(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Completion:
        pass

    @abstractmethod
    async def generate_completion_stream(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[CompletionChunk]:
        pass

    @abstractmethod
    async def generate_message_batch(
        self,
        message_history: MessageHistory,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Sequence[AssistantMessage]:
        pass
